


#include  <stdio.h>			
#include  <stdlib.h>			
#include  <string.h>			
#include  <pthread.h>			

#include  "redundancy.h"			
#include  "returncode.h"			





int pthread_redundancy_attr_destroy(struct pthread_redundancy_attr_t *a) {
   return 0;
}


int pthread_redundancy_attr_init(struct pthread_redundancy_attr_t *a) {
   a->size=1;
   a->priority=0;
   return 0;
}


int pthread_redundancy_attr_setpriority(struct pthread_redundancy_attr_t *a, int prio) {
   a->priority=prio;
   return 0;
}


int pthread_redundancy_attr_getpriority(struct pthread_redundancy_attr_t a) {
  return a.priority;
}

int pthread_redundancy_attr_setsize(struct pthread_redundancy_attr_t *a, int s) {
   a->size=s;
   return 0;
}

int pthread_redundancy_attr_getsize(struct pthread_redundancy_attr_t a) {
  return a.size;
}



typedef struct redundancy_thread_argument_t {
	int ego;
	struct pthread_redundancy_t* a_redundancy;
} redundancy_thread_argument_t;



void* redundancy_thread(void* arg) {
  struct redundancy_thread_argument_t* a = (redundancy_thread_argument_t*)arg;
  int status;

  while(1) {

    status=sem_wait(&a->a_redundancy->waitfor_barrier[a->ego]);
    if(status!=0)
        {
           returncode("sem_wait redundancy_thread waitfor",status);
           pthread_exit(NULL);
        }        

    if (a->a_redundancy->cont==0) 
        pthread_exit(NULL);

    a->a_redundancy->results[a->ego]=a->a_redundancy->voters[a->ego](a->a_redundancy->arg);

    status=sem_post(&a->a_redundancy->completion_barrier[a->ego]);
    if(status!=0)
        {
           returncode("sem_post redundancy_thread completion",status);
           pthread_exit(NULL);
        }        

  }
  return NULL;
}




int pthread_redundancy_init(struct pthread_redundancy_t* r, struct pthread_redundancy_attr_t a, voter_type v [], void* arg) {
  
  int status;
  int i;
  pthread_t tid;
  pthread_attr_t attr;
  struct sched_param param;
  struct redundancy_thread_argument_t* a_thread_arg;


  r->attr=a;
  r->arg=arg;
  r->cont=1;
  for(i=0; i<r->attr.size; i++) 
    {
    r->voters[i]=v[i];

    status=sem_init(&r->waitfor_barrier[i],0,0);
    if(status!=0)
        {
           returncode("pthread_redundancy_init sem_init waitforbarrier",status);
           return status;
        }        

    status=sem_init(&r->completion_barrier[i],0,0);
    if(status!=0)
        {
           returncode("pthread_redundancy_init sem_init completion_barrier",status);
           return status;
        }        
    }


  pthread_attr_init(&attr);
  status=pthread_attr_setinheritsched(&attr,PTHREAD_EXPLICIT_SCHED);
  if(status<0)
      {
         returncode("pthread_redundancy_init pthread_attr_setinheritsched",status);
         return status;
      }        
  status=pthread_attr_setschedpolicy(&attr,SCHED_FIFO);
  if(status<0)
      {
         returncode("pthread_redundancy_init pthread_attr_setschedpolicy",status);
         return status;
      }        
  param.sched_priority=r->attr.priority;
  status=pthread_attr_setschedparam(&attr,&param);
  if(status<0)
      {
         returncode("pthread_redundancy_init pthread_attr_setschedparam",status);
         return status;
      }        

  for(i=0; i<r->attr.size; i++) 
    {
    /* allocate thread argument */
    a_thread_arg = (struct redundancy_thread_argument_t*)malloc(sizeof(struct redundancy_thread_argument_t));
    a_thread_arg->ego=i;
    a_thread_arg->a_redundancy=r;

    /* create thread */
    status=pthread_create(&tid,&attr,redundancy_thread,a_thread_arg);
    if(status<0)
        {
           returncode("pthread_redundancy_init pthread_create",status);
           return status;
        }        
    }


    return 0;
}

int pthread_redundancy_vote(struct pthread_redundancy_t* r, int* result) {

  int status, i, j;
  int maxcount;

  for(i=0; i<r->attr.size; i++) 
    {
    status=sem_post(&r->waitfor_barrier[i]);
    if(status!=0)
        {
           returncode("pthread_redundancy_vote sem_post waitfor",status);
           return status;
        }        
    }

  for(i=0; i<r->attr.size; i++) 
    {
    status=sem_wait(&r->completion_barrier[i]);
    if(status!=0)
        {
           returncode("pthread_redundancy_vote sem_wait completion",status);
           return status;
        }        
    }

  maxcount=0;
  for (i = 0; i < r->attr.size; i++) {
        int count = 0;
        for (j = 0; j < r->attr.size; j++) {
            if (r->results[i] == r->results[j])
                count++;
        }
 
        if (count > maxcount) {
            maxcount = count;
            *result = r->results[i];
        }
  }
 
  return r->attr.size - maxcount;
}


int pthread_redundancy_destroy(struct pthread_redundancy_t* r) {

  int i;
  int status;

  r->cont=0;
  for(i=0; i<r->attr.size; i++) 
    {
    status=sem_post(&r->waitfor_barrier[i]);
    if(status!=0)
        {
           returncode("pthread_redundancy_destroy sem_post",status);
           return status;
        }        
    }

  return 0;
}






